

--
-- Copyright (C) 2016  <fastrgv@gmail.com>
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You may read the full text of the GNU General Public License
-- at <http://www.gnu.org/licenses/>.
--






-- seven.adb
-- 23jun16
--
-- retro-text version of flat7 intended to run in a terminal
-- window using Ada to facilitate character input from keyboard.
--
-- Build instructions for all platforms:
-- Manually install GNAT GPL from libre.adacore.com/download/.
-- then type "gnatmake seven"





with Ada.Text_IO; use Ada.Text_IO;
with ada.numerics.generic_elementary_functions;
with Ada.Command_Line;
with SysUtils;  use SysUtils;
with Ada.Numerics.Float_Random;

with GNATCOLL.Terminal;  use GNATCOLL.Terminal;



procedure seven is


	G : Ada.Numerics.Float_Random.Generator;

	ch: character;


userexit : boolean := false;


dimen : constant integer := 2;
ncubes  : constant integer := dimen*dimen*dimen;

subtype rngs is integer range 1..dimen;
subtype rngm is integer range 1..ncubes;
type permtype is array(rngs,rngs,rngs) of rngm;

perm : permtype;
brow, bcol, blay : rngs := 2;
winner, help : boolean := false;


subtype str1 is string(1..1);
symbol : constant array(1..ncubes) of str1
     := ( "1", "2", "3",   "4", "5", "6",   "7", " " );





function indx(row,col,lay:rngs) return rngm is
begin
	return (lay-1)*dimen*dimen + (row-1)*dimen + col;
end indx;





procedure test4winner is
	k : integer := 1;
begin
	winner := (brow=dimen) and (bcol=dimen) and (blay=dimen);
	for lay in rngS loop
	for row in rngS loop
	for col in rngS loop
		winner := winner and (perm(row,col,lay) = k);
		k:=k+1;
	end loop;
	end loop;
	end loop;
end test4winner;











   procedure movenear is
   -- move space away (to a bigger layer#)
   begin
     if blay<dimen then
       perm(brow,bcol,blay):=perm(brow,bcol,blay+1);
       blay:=blay+1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end movenear;

   procedure moveaway is
   -- move space closer (to a smaller layer#)
   begin
     if blay>1 then
       perm(brow,bcol,blay):=perm(brow,bcol,blay-1);
       blay:=blay-1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end moveaway;





   procedure moveleft is
   --move space to a bigger col#
   begin
     if bcol<dimen then
       perm(brow,bcol,blay):=perm(brow,bcol+1,blay);
       bcol:=bcol+1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end moveleft;

   procedure moveright is
   --move space to a smaller col#
   begin
     if bcol>1 then
       perm(brow,bcol,blay):=perm(brow,bcol-1,blay);
       bcol:=bcol-1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end moveright;










   procedure movedown is
   --move space to a smaller row#
   begin
     if brow>1 then
       perm(brow,bcol,blay):=perm(brow-1,bcol,blay);
       brow:=brow-1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end movedown;

   procedure moveup is
   --move space to a bigger row#
   begin
     if brow<dimen then
       perm(brow,bcol,blay):=perm(brow+1,bcol,blay);
       brow:=brow+1;
       perm(brow,bcol,blay):=ncubes;
		 test4winner;

     end if;
   end moveup;















up : constant str1 := "u";
dn : constant str1 := "d";
lt : constant str1 := "l";
rt : constant str1 := "r";
nr : constant str1 := "n";
aw : constant str1 := "a";
sol : array(1..100_000) of str1;
nsol : integer := 0;






   procedure shuffle( level: in integer ) is

	  r : Ada.Numerics.Float_Random.Uniformly_Distributed;

     n : integer;
	  br1,br2,bc1,bc2,bl1,bl2: integer;
	  prev : str1 := "z";
   begin

		winner:=false;
     for row in rngS loop
       for col in rngS loop
         for lay in rngS loop
           perm(row,col,lay) := indx(row,col,lay);
	 		end loop;
       end loop;
     end loop;
     brow:=dimen;
     bcol:=dimen;
     blay:=dimen;

     case level is
     when 0 => n:=0;
     when 1 => n:=10;
     when 2 => n:=100;
     when 3 => n:=1000;
     when 4 => n:=10_000;
     when 5 => n:=100_000;
     when others => n:=0;
     end case;


-- here we use a "stack" of letters {l,r,u,d,n,a} to 
-- store solution for possible playback
-- WARNING:  for large n, stored solution is 
--           likely very far from optimal!

	  nsol:=0;
     for i in 1..n loop

		 r := Ada.Numerics.Float_Random.Random(G);

       if r<0.16 and prev /= rt then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         moveleft;                                --attempted move
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then --attempt succeeded
				prev:=lt;
				nsol:=nsol+1; sol(nsol):=rt;          --store move
			end if;
       elsif r<0.33 and prev /= lt then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         moveright;											--attempt
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then  --success
				prev:=rt;
				nsol:=nsol+1; sol(nsol):=lt;           --store
			end if;
       elsif r<0.49 and prev /= dn then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         moveup;
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then
				prev:=up;
				nsol:=nsol+1; sol(nsol):=dn;
			end if;
       elsif r<0.66 and prev /= up then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         movedown;
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then
				prev:=dn;
				nsol:=nsol+1; sol(nsol):=up;
			end if;
       elsif r<0.82 and prev /= aw then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         movenear;
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then
				prev:=nr;
				nsol:=nsol+1; sol(nsol):=aw;
			end if;
       elsif prev /= nr then
		 	br1:=brow; bc1:=bcol; bl1:=blay;
         moveaway;
		 	br2:=brow; bc2:=bcol; bl2:=blay;
			if br2/=br1 or bc2/=bc1 or bl2/=bl1 then
				prev:=aw;
				nsol:=nsol+1; sol(nsol):=nr;
			end if;
       end if;

     end loop;
   end shuffle;








procedure initperm is
begin

   for row in rngS loop
     for col in rngS loop
       for lay in rngS loop
           perm(row,col,lay) := indx(row,col,lay);
       end loop;
     end loop;
   end loop;
	brow:=dimen;
	bcol:=dimen;
	blay:=dimen;
	if perm(brow,bcol,blay) /= ncubes then
		raise program_error;
	end if;
end initperm;











procedure Draw is
	Ok: boolean;
	s : str1;
	info: terminal_info;
begin

	info.init_for_stdout(auto);

	SysUtils.Shell("clear", Ok); -- erase-terminal

if help then

	put_line(" h,?=toggle-help,  key-map:");
	put_line("=====================================");
	put_line(" layer up = LU:");
	put_line(" <u>, <.>, <\>, <home>, <->");
	new_line;
	put_line(" layer down = LD:");
	put_line("  <o>, </>, <end>, <+>");
	put_line("=====================================");
	put_line(" old-school controls:");
	put_line(" u=LU   i=no   o=LD");
	put_line(" j=we   k=so   l=ea");
	put_line("=====================================");
	put_line(" modern controls:");
	put_line(" use arrow keys within a layer,");
	put_line(" (+), (-) keys move between layers.");
	put_line("=====================================");

else

	put_line(" 3D Seven Slider...Scramble then restore ordering");
	put_line(" q=quit,  ?=toggle-help,  1..5 to scramble");
	new_line;

   Info.Set_Color (style=>bright);
   Info.Set_Color (background=>black);

	info.set_color(foreground=>magenta);

	for lay in 1..dimen loop

		if lay=blay then
		   Info.Set_Color (background=>blue);
		else
		   Info.Set_Color (background=>black);
		end if;

		if lay=1 then
			info.set_color(foreground=>green);
		elsif lay=2 then
			info.set_color(foreground=>cyan);
		end if;

		put_line(" ----- ");
		for row in 1..dimen loop
			put("|");
			for col in 1..dimen loop
				s := symbol( perm(row,col,lay) );
				put(' '&s);
			end loop;
			put_line(" |");
		end loop;
	end loop;

	put_line(" ----- ");

	if winner then
	   Info.Set_Color (background=>black);
		info.set_color(foreground=>yellow);
		put_line(" Correct !");
	end if;

   Info.Set_Color (Standard_Output, Style => Reset_All);

end if;

end Draw;










procedure handle_key_down( ch: character ) is
begin

-- note:  <home>='H', <end>='F', <up>='A', etc

	case ch is

		when '1' => shuffle(1);
		when '2' => shuffle(2);
		when '3' => shuffle(3);
		when '4' => shuffle(4);
		when '5' => shuffle(5);

		when 'q' | 'x' =>	userexit:=true;

		when 'h' | '?' => help := not help;

		when 'A' | 'i' | 'w' =>	moveup;

		when 'B' | 'k' | 's' =>	movedown;

		when 'C' | 'l' | 'd' =>	moveright;

		when 'D' | 'j' | 'a' =>	moveleft;

		-- layer-down <end>=>'F'
		when 'F' | 'o' | '/' | '+' => moveaway;

		-- layer-up <home>=>'H'
		when 'H' | 'u' | '.' | '\' | '-' => movenear;

		-- this is immediately erased unless
		-- we pause for user input...
		when others => null;

	end case;

end handle_key_down;











begin --seven


	InitPerm;
	Ada.Numerics.Float_Random.Reset(G); --randomize (time-dependent)

	Draw;

	while not userexit loop
		get_immediate(ch);
		handle_key_down( ch );
		Draw;
	end loop;


end seven;






